#include <cstdint>
#include <iostream>


// Functions of the BitMatrix
uint64_t *createMatrix(int rows, int cols);
void destroyMatrix(uint64_t *matrix);
void setBit(uint64_t *matrix, int total_cols, int row, int col);
void clearBit(uint64_t *matrix, int total_cols, int row, int col);
void toggleBit(uint64_t *matrix, int total_cols, int row, int col);
bool testBit(uint64_t *matrix, int total_cols, int row, int col);
void printMatrix(uint64_t *matrix, int total_rows, int total_cols);

// Testing functions
bool testMatrix(int rows, int cols);
bool testMatrixEqual(uint64_t *matrix, bool **oracle, int total_rows, int total_cols);
void printOracle(bool **oracle, int total_rows, int total_cols);


uint64_t *createMatrix(int rows, int cols) {
  // Calculate number of uint64_ts needed
  int size = (rows * cols + 63) / 64;
  // Allocate space
  uint64_t *matrix = new uint64_t[size];
  // Initialize the space to all zeros
  for (int i = 0; i < size; ++i)
    matrix[i] = uint64_t{0};

  return matrix;
}

void destroyMatrix(uint64_t *matrix) {
  delete[] matrix;
}

void setBit(uint64_t *matrix, int total_cols, int row, int col) {
  int bit = row * total_cols + col;
  int block = bit / 64;
  int offset = bit % 64;

  uint64_t mask = uint64_t{1} << offset;
  matrix[block] |= mask;
}

void clearBit(uint64_t *matrix, int total_cols, int row, int col) {
  int bit = row * total_cols + col;
  int block = bit / 64;
  int offset = bit % 64;

  uint64_t mask = ~(uint64_t{1} << offset);
  matrix[block] &= mask;
}

void toggleBit(uint64_t *matrix, int total_cols, int row, int col) {
  int bit = row * total_cols + col;
  int block = bit / 64;
  int offset = bit % 64;

  uint64_t mask = uint64_t{1} << offset;
  matrix[block] ^= mask;
}

bool testBit(uint64_t *matrix, int total_cols, int row, int col) {
  int bit = row * total_cols + col;
  int block = bit / 64;
  int offset = bit % 64;

  uint64_t mask = uint64_t{1} << offset;
  return (matrix[block] & mask) > 0;
}

void printMatrix(uint64_t *matrix, int total_rows, int total_cols) {
  for (int i = 0; i < total_rows; ++i) {
    for (int j = 0; j < total_cols; ++j) {
      if (testBit(matrix, total_cols, i, j))
        std::cout << "1 ";
      else
        std::cout << "0 ";
    }
    std::cout << "\n";
  }
  std::cout << "\n";
}


bool testMatrix(int rows, int cols) {
  std::cerr << "Start test for " << rows << "x" << cols << "\n";
  // Create matrix
  uint64_t *matrix = createMatrix(rows, cols);

  // Create oracle
  bool **oracle = new bool*[rows];
  for (int i = 0; i < rows; ++i)
    oracle[i] = new bool[cols];
  // Initialize oracle to all false to match matrix
  for (int i = 0; i < rows; ++i) {
    for (int j = 0; j < cols; ++j) {
      oracle[i][j] = false;
    }
  }

  // Testing code
  bool pass = true;

  // Ensure they match to start
  if (!testMatrixEqual(matrix, oracle, rows, cols)) {
    pass = false;
    std::cerr << "Failed initial matrix equal to ensure they match at start.";
  }

  // Set bits in row-major order
  for (int i = 0; i < rows; ++i) {
    for (int j = 0; j < cols; ++j) {
      setBit(matrix, cols, i, j);
      oracle[i][j] = true;
      if (!testMatrixEqual(matrix, oracle, rows, cols)) {
        std::cerr << "setBit (" << i << ", " << j << ")\n";
      }
    }
  }

  // Clear bits in row-major order
  for (int i = 0; i < rows; ++i) {
    for (int j = 0; j < cols; ++j) {
      clearBit(matrix, cols, i, j);
      oracle[i][j] = false;
      if (!testMatrixEqual(matrix, oracle, rows, cols)) {
        std::cerr << "clearBit (" << i << ", " << j << ")\n";
      }
    }
  }

  // Set bits in reverse row-major order
  for (int i = rows - 1; i >= 0; --i) {
    for (int j = cols - 1; j >= 0; --j) {
      setBit(matrix, cols, i, j);
      oracle[i][j] = true;
      if (!testMatrixEqual(matrix, oracle, rows, cols)) {
        std::cerr << "setBit reverse (" << i << ", " << j << ")\n";
      }
    }
  }

  // Clear bits in reverse row-major order
  for (int i = rows - 1; i >= 0; --i) {
    for (int j = cols - 1; j >= 0; --j) {
      clearBit(matrix, cols, i, j);
      oracle[i][j] = false;
      if (!testMatrixEqual(matrix, oracle, rows, cols)) {
        std::cerr << "clearBit reverse (" << i << ", " << j << ")\n";
      }
    }
  }

  // Toggle bits in row-major order
  for (int i = 0; i < rows; ++i) {
    for (int j = 0; j < cols; ++j) {
      toggleBit(matrix, cols, i, j);
      oracle[i][j] ^= true;
      if (!testMatrixEqual(matrix, oracle, rows, cols)) {
        std::cerr << "setBit (" << i << ", " << j << ")\n";
      }
    }
  }

  // Clean up oracle
  for (int i = 0; i < rows; ++i)
    delete[] oracle[i];
  delete[] oracle;

  // Clean up matrix
  destroyMatrix(matrix);

  std::cerr << "Finished test for " << rows << "x" << cols;
  if (pass)
    std::cerr << " passed.\n";
  else
    std::cerr << " failed.\n";

  return pass;
}

bool testMatrixEqual(uint64_t *matrix, bool **oracle, int total_rows, int total_cols) {
  for (int i = 0; i < total_rows; ++i) {
    for (int j = 0; j < total_cols; ++j) {
      if (oracle[i][j] != testBit(matrix, total_cols, i, j)) {
        std::cerr << "Fail equal at (" << i << ", " << j << ")\n";
        std::cerr << "Bit Matrix:\n";
        printMatrix(matrix, total_rows, total_cols);
        std::cerr << "Oracle:\n";
        printOracle(oracle, total_rows, total_cols);
        return false;
      }
    }
  }
  return true;
}

void printOracle(bool **oracle, int total_rows, int total_cols) {
  for (int i = 0; i < total_rows; ++i) {
    for (int j = 0; j < total_cols; ++j) {
      if (oracle[i][j])
        std::cout << "1 ";
      else
        std::cout << "0 ";
    }
    std::cout << "\n";
  }
  std::cout << "\n";
}


int main() {

  testMatrix(10, 10);
  testMatrix(100, 100);

  testMatrix(8, 8);
  testMatrix(4, 16);
  testMatrix(16, 4);
  testMatrix(1, 64);
  testMatrix(1, 65);
  testMatrix(64, 1);
  testMatrix(65, 1);
  testMatrix(64, 64);
  testMatrix(64, 65);
  testMatrix(65, 64);
  testMatrix(65, 65);


  return 0;
}
